/* Copyright INRIA 2004
**
** This file is part of the Scotch distribution.
**
** The Scotch distribution is libre/free software; you can
** redistribute it and/or modify it under the terms of the
** GNU Lesser General Public License as published by the
** Free Software Foundation; either version 2.1 of the
** License, or (at your option) any later version.
**
** The Scotch distribution is distributed in the hope that
** it will be useful, but WITHOUT ANY WARRANTY; without even
** the implied warranty of MERCHANTABILITY or FITNESS FOR A
** PARTICULAR PURPOSE. See the GNU Lesser General Public
** License for more details.
**
** You should have received a copy of the GNU Lesser General
** Public License along with the Scotch distribution; if not,
** write to the Free Software Foundation, Inc.,
** 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
**
** $Id$
*/
/************************************************************/
/**                                                        **/
/**   NAME       : hmesh_order_hx.c                        **/
/**                                                        **/
/**   AUTHOR     : Francois PELLEGRINI                     **/
/**                                                        **/
/**   FUNCTION   : This module contains service routines   **/
/**                for the hmeshOrderH{d|f} ordering       **/
/**                routines.                               **/
/**                                                        **/
/**   DATES      : # Version 4.0  : from : 09 dec 2003     **/
/**                                 to   : 24 jan 2004     **/
/**                                                        **/
/************************************************************/

/*
**  The defines and includes.
*/

#define HMESH_ORDER_HX

#include "module.h"
#include "common.h"
#include "graph.h"
#include "mesh.h"
#include "hmesh.h"
#include "hmesh_order_hx.h"

/***********************************/
/*                                 */
/* These are the service routines. */
/*                                 */
/***********************************/

/* This routine fills the input arrays for
** the mesh ordering routines.
** It returns:
** - void  : in all cases.
*/

int
hmeshOrderHxFill (
const Hmesh * restrict const              meshptr,
Gnum * restrict const                     petab,
Gnum * restrict const                     lentab,
Gnum * restrict const                     iwtab,
Gnum * restrict const                     elentab,
Gnum * restrict const                     pfreptr)
{
  Gnum * restrict             petax;
  Gnum * restrict             iwtax;
  Gnum * restrict             lentax;
  Gnum * restrict             elentax;
  HmeshOrderHxHash * restrict hashtab;            /* Neighbor hash table */
  Gnum                        hashsiz;
  Gnum                        hashmsk;
  Gnum                        n;                  /* Number of nodes to order             */
  Gnum                        velmadj;            /* Index adjustment for element indices */
  Gnum                        vnodadj;            /* Index adjustment for node indices    */
  Gnum                        velmnum;
  Gnum                        vnodnum;
  Gnum                        degrval;
  Gnum                        vertnum;
  Gnum                        edgenum;

  n = meshptr->m.velmnbr + meshptr->m.vnodnbr;
  for (hashsiz = 16, degrval = meshptr->m.degrmax * (meshptr->m.degrmax - 1); /* Compute hash table size */
       hashsiz < degrval; hashsiz <<= 1) ;
  hashsiz <<= 1;
  hashmsk   = hashsiz - 1;

  if ((hashtab = memAlloc (hashsiz * sizeof (HmeshOrderHxHash))) == NULL) {
    errorPrint  ("hmeshOrderHxFill: out of memory");
    return      (1);
  }
  memSet (hashtab, ~0, hashsiz * sizeof (HmeshOrderHxHash));

  petax   = petab - 1;                            /* Base HAMF arrays at base 1 */
  iwtax   = iwtab - 1;
  lentax  = lentab - 1;
  elentax = elentab - 1;

  velmadj = 1 + meshptr->m.vnodnbr - meshptr->m.velmbas;
  for (vnodnum = meshptr->m.vnodbas, vertnum = edgenum = 1; /* Copy non-halo node data with base 1 */
       vnodnum < meshptr->vnohnnd; vertnum ++, vnodnum ++) {
    Gnum                      enodnum;
    Gnum                      nghbnbr;

    petax[vertnum]  = edgenum;
    lentax[vertnum] = meshptr->m.vendtax[vnodnum] - meshptr->m.verttax[vnodnum];

    for (enodnum = meshptr->m.verttax[vnodnum], nghbnbr = -1; /* -1 since loop edge will be processed in the main loop */
         enodnum < meshptr->m.vendtax[vnodnum]; enodnum ++) {
      Gnum                      velmnum;
      Gnum                      eelmnum;

      velmnum = meshptr->m.edgetax[enodnum];

      iwtax[edgenum ++] = velmnum + velmadj;      /* Adjust end element index */
      for (eelmnum = meshptr->m.verttax[velmnum]; eelmnum < meshptr->m.vendtax[velmnum]; eelmnum ++) {
        Gnum                      vnodend;
        Gnum                      hnodend;

        vnodend = meshptr->m.edgetax[eelmnum];

        for (hnodend = (vnodend * HMESHORDERHXHASHPRIME) & hashmsk; ; hnodend = (hnodend + 1) & hashmsk) {
          if (hashtab[hnodend].vertnum != vnodnum) {
            hashtab[hnodend].vertnum = vnodnum;
            hashtab[hnodend].vertend = vnodend;
            nghbnbr ++;
          }
          if (hashtab[hnodend].vertend == vnodend) /* If end vertex already present */
            break;                                /* Skip to next end vertex        */
        }
      }
      elentax[vertnum] = nghbnbr;
    }
  }

  for ( ; vnodnum < meshptr->m.vnodnnd; vnodnum ++, vertnum ++) { /* Copy halo vertices with base 1 */
    Gnum                      degrval;
    Gnum                      enodnum;

    degrval = meshptr->m.verttax[vnodnum] - meshptr->m.vendtax[vnodnum];
    petax[vertnum]   = edgenum;
    lentax[vertnum]  = (degrval != 0) ? degrval : - (n + 1);
    elentax[vertnum] = 0;

    for (enodnum = meshptr->m.verttax[vnodnum];
         enodnum < meshptr->m.vendtax[vnodnum]; enodnum ++)
      iwtax[edgenum ++] = meshptr->m.edgetax[enodnum] + velmadj; /* Adjust end element index */
  }

  vnodadj = 1 - meshptr->m.vnodbas;               /* Base nodes at 1 */
  for (velmnum = meshptr->m.velmbas; velmnum < meshptr->m.velmnnd; velmnum ++, vertnum ++) {
    Gnum                      eelmnum;

    petax[vertnum]   = edgenum;
    lentax[vertnum]  = meshptr->m.vendtax[velmnum] - meshptr->m.verttax[velmnum];
    elentax[vertnum] = - (n + 1);

    for (eelmnum = meshptr->m.verttax[velmnum];
         eelmnum < meshptr->m.vendtax[velmnum]; eelmnum ++)
      iwtax[edgenum ++] = meshptr->m.edgetax[eelmnum] + vnodadj; /* Adjust end node index */
  }

  *pfreptr = edgenum;                             /* Set index to first free area */

  memFree (hashtab);

  return (0);
}
